# -*- coding: utf-8 -*-

"""Config object for hreports."""

import os
import copy
import yaml
import click


APP_NAME = 'hreports'


class Config(object):

    def __init__(self):
        self.verbose = False
        self.data = {'global': {}, 'reports': {}}
        self.read_config()

    def get_stored_reports(self):
        return self.data.get('reports')

    def store_report_data(self, name, data):
        if name not in self.data.get('reports').keys():
            self.data.get('reports')[name] = {}

        # Remove none value
        data_dict = {}
        for key, value in data.items():
            if value:
                data_dict[key] = value

        self.data.get('reports')[name].update(data_dict)

    def delete_report(self, name):
        if name in self.data.get('reports').keys():
            self.data.get('reports').pop(name)
        self.write_config()

    def copy_report(self, source, target):
        source_config = self.data.get('reports').get(source)
        click.echo(source_config)
        if source_config:
            target_config = copy.deepcopy(source_config)
            self.data.get('reports')[target] = target_config
        self.write_config()

    def update_report(self, name, meta=False, variables=False, write=True):
        if meta:
            self.store_report_data(name, meta)
        if variables:
            report_config = self.data.get('reports').get(name)
            section_dict = report_config.get('variables', {})

            # Click collects --variables data as a list tuples
            for item in variables:
                section_dict[item[0]] = item[1]
            self.store_report_data(name, {'variables': section_dict})

        if write:
            self.write_config()

    def read_config(self):
        self.cfg_file = os.path.join(click.get_app_dir(APP_NAME),
                                     'config.yaml')
        try:
            self.data = yaml.safe_load(open(self.cfg_file))
        except yaml.YAMLError as exc:
            click.echo(exc)
        except EnvironmentError:
            if self.verbose:
                click.echo('Config file not found. Creating one.')
            self.write_config()
        if self.verbose:
            click.echo('Loaded config file %s' % self.cfg_file)

    def write_config(self):

        config_dir = os.path.dirname(self.cfg_file)

        if not os.path.exists(config_dir):
                os.makedirs(config_dir)

        with open(self.cfg_file, 'w') as outfile:
            yaml.safe_dump(self.data, outfile,
                           encoding='utf-8',
                           default_flow_style=False,
                           allow_unicode=True)
        if self.verbose:
            click.echo('Config file updated at %s.' % self.cfg_file)

    def echo_saved_reports(self):
        if not self.get_stored_reports():
            click.secho('No reports saved.', fg='red')
            return
        click.echo('Stored reports:')
        report_names = self.get_stored_reports().keys()
        report_names.sort()
        report_name_length = max([len(name) for name in report_names])

        for report_name in report_names:
            cfg = self.get_stored_reports().get(report_name)
            if cfg.get('desc', None):
                click.secho(report_name.ljust(report_name_length),
                            fg='green', nl=False)
                click.secho(' -- %s' % cfg.get('desc'),
                            )
            else:
                click.secho('%s' % report_name, fg='green')



# -*- coding: utf-8 -*-

"""Console script for hreports."""

import os
import click
import yaml

import hreports
from .config import Config


def composed(*decs):
    """Combine multiple decorators.
    See https://stackoverflow.com/questions/5409450/
    """

    def deco(f):
        for dec in reversed(decs):
            f = dec(f)
        return f
    return deco


common = composed(click.option('--query', '-q', required=False),
                  click.option('--template', '-t', required=False),
                  click.option('--filename', '-f', required=False),
                  click.option('--ledger', '-l', required=False),
                  click.option('--desc', '-d', required=False),
                  click.option('--variables', '-var', required=False,
                               type=(str, str), multiple=True)
                  )


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(invoke_without_command=True,
             context_settings=CONTEXT_SETTINGS)
@click.option('--config-file', '-c', required=False,
              type=click.Path(exists=True),
              help='Use config file')
@click.option('--config-info', '-i', is_flag=True,
              help='Show current config file')
@click.option('--verbose', is_flag=True)
@click.option('--query', '-q', required=False,
              metavar='QUERY',
              help='Execute QUERY query')
@click.option('--report-config', '-r', required=False,
              metavar='REPORT',
              help='Show configuration of REPORT')
@click.option('--ledger', '-l', metavar='FILE', required=False,
              type=click.Path(exists=True),
              help='Use FILE ledger')
@click.version_option()
@click.pass_context
def main(context, config_info, config_file, ledger, verbose,
         query, report_config):
    """Manage hledger queries."""
    config = Config()
    context.obj = config
    click.echo(ledger)
    context.obj.ledger = ledger

    if verbose:
        config.verbose = True
        click.echo("Config file %s" % config.cfg_file)
    if not context.invoked_subcommand and query:
        hreport = hreports.Hreport(config)
        click.echo("Running %s" % query)
        click.echo(hreport.run(query=query))
    elif report_config:
        report_config = config.get_stored_reports().get(report_config)
        click.echo(yaml.dump(report_config, default_flow_style=False))
    elif config_info:
        print(yaml.dump(config.data))
    elif not context.invoked_subcommand:
        config.echo_saved_reports()


@main.command(short_help='Create a report')
@click.argument('name')
@common
@click.pass_obj
def create(config, name, variables, **meta):
    click.echo('Saving %s report' % name)
    if name in config.get_stored_reports():
        click.confirm('Do you want to overwrite %s?' % name,
                      abort=True)
        config.delete_report(name)
    config.update_report(name, meta, variables)


@main.command(short_help='Update report configuration')
@click.argument('name')
@common
@click.pass_obj
def update(config, name, variables, **meta):
    click.echo('Updating  %s report' % name)
    config.update_report(name, meta, variables)


@main.command(short_help='Delete report')
@click.argument('name')
@click.pass_obj
def delete(config, name):
    if name not in config.get_stored_reports():
        raise click.UsageError('Report %s does not exist.' % name)
    else:
        click.echo('Deleting %s report' % name)
        config.delete_report(name)


@main.command(short_help='Show report result')
@click.argument('name', required=False)
@common
@click.pass_obj
def show(config, name, variables, **meta):

    if config.verbose:
        click.echo("Showing reports", nl=True)

    if not name or name not in config.get_stored_reports():
        config.echo_saved_reports()
    elif name:
        config.update_report(name, meta=meta, variables=variables,
                             write=False)
        hreport = hreports.Hreport(config)
        click.echo(hreport.render(name))

        if config.verbose:
            click.echo('Ran query "%s"' % config.cmd)


@main.command(short_help='Copy report configuration')
@click.argument('source')
@click.argument('target')
@click.pass_obj
def copy(config, source, target):
    source_config = config.get_stored_reports().get(source, None)

    if not source_config:
        raise click.UsageError('Report %s does not exist.' % source)

    if target in config.get_stored_reports():
        click.confirm('Do you want to overwrite %s?' % target,
                      abort=True)

    config.copy_report(source, target)


@main.command(short_help='Edit configuration file or template')
@click.option('--template', '-t', required=False)
@click.pass_obj
def edit(config, template):
    if template:
        template_file = os.path.join(click.get_app_dir('hreports'),
                                     'templates',
                                     template)

        if not os.path.exists(os.path.dirname(template_file)):
                os.makedirs(os.path.dirname(template_file))

        click.edit(filename=template_file)
    else:
        click.edit(filename=config.cfg_file)


@click.argument('name')
@main.command(short_help='Save report to pdf file')
@common
@click.option('--file-type', '-f', required=False,
              type=click.Choice(['md', 'pdf', 'html']))
@click.pass_obj
def save(config, name, variables, **meta):
    config.update_report(name, meta, variables)
    hreport = hreports.Hreport(config)
    output_file = hreport.save(name)
    click.echo('Saved %s' % output_file)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Main module."""

import os
import io
import subprocess
import datetime
import tempfile
import shlex
from jinja2 import Environment, ChoiceLoader, \
    FileSystemLoader, PackageLoader, select_autoescape
from jinja2.exceptions import TemplateSyntaxError, TemplateNotFound, \
    UndefinedError
from .template_filters import datetimeformat, german_float, \
    last_day_of_month, substract_days, multiply_last_column, \
    add_percentage_column


class Hreport(object):
    def __init__(self, config):
        self.config = config

        cfg_path = os.path.dirname(self.config.cfg_file)
        self.cfg_templates = os.path.join(cfg_path, 'templates')

        loader = ChoiceLoader([
                    FileSystemLoader('./'),
                    FileSystemLoader('/templates'),
                    FileSystemLoader(self.cfg_templates),
                    PackageLoader('hreports', 'templates'),
        ])
        self.env = Environment(loader=loader,
                               autoescape=select_autoescape(['html', 'xml'])
                               )

        self.env.filters['datetime'] = datetimeformat
        self.env.filters['german_float'] = german_float
        self.env.filters['last_day_of_month'] = last_day_of_month
        self.env.filters['substract_days'] = substract_days
        self.env.filters['multiply_last_column'] = multiply_last_column
        self.env.filters['add_percentage_column'] = add_percentage_column

    def get_global_config(self):
        return self.config.data.get('global', None)

    def get_report_config(self, name):
        return self.config.data.get('reports').get(name, None)

    def get_report_config_value(self, name, key):
        report_config = self.config.data.get('reports').get(name, None)
        if report_config:
            value = report_config.get(key, None)
        else:
            value = False
        if not value:
            # Try global section
            return self.get_global_config_value(value)
        return value

    def get_global_config_value(self, key):
        return self.config.data.get('global').get(key, None)

    def run(self, name=False, query=False, ledger=False):
        if name:
            query = self.get_report_config_value(name, 'query')
            ledger = self.get_report_config_value(name, 'ledger')

        if not ledger:
            ledger = self.get_global_config_value('ledger')

        query = self.render_string(query, name)

        if ledger:
            cmd = 'hledger -f %s %s' % (ledger, query)
        else:
            cmd = 'hledger %s' % query
        self.config.cmd = cmd
        try:
            cmd_list = shlex.split(cmd)
            output = unicode(subprocess.check_output(cmd_list), 'utf-8')
            self.config.returncode = 0
        except subprocess.CalledProcessError as exception:
            self.config.error = exception.output
            self.config.returncode = exception.returncode
            output = 'Query %s returned non-zero exit status' % cmd
        return output

    def render_string(self, string, name=False):
        string_template = self.env.from_string(string)
        try:
            string = string_template.render(self.get_context(name))
        except TemplateSyntaxError:
            string = 'Encountered syntax error in %s' % string
        except UndefinedError as exception:
            string = 'Encountered variable UndefinedError: %s' % \
                exception.message
        return string

    def render_strings_in_dict(self, data_dict, context, section=False):
        if not data_dict:
            return context
        for key, value in data_dict.items():
            if isinstance(value, str):
                value_template = self.env.from_string(value)
                data_dict[key] = value_template.render(context)
        if section:
            context.update({section: data_dict})
        else:
            context.update(data_dict)
        return context

    def get_context(self, name=False):
        """Build context data for the template.

        The template context dict is evaluated in this order:

            1. Builtins at {}
            2. variables of the global section at {}
            3. global section at {'global'}
            4. variables of the report section {'report'}
            5 report section at {}
        """
        context = {}

        builtins = {'now': datetime.datetime.now()}
        context.update(builtins)

        global_config = self.get_global_config()
        if 'variables' in global_config:
            global_variables = global_config.get('variables', {})
            context = self.render_strings_in_dict(global_variables,
                                                  context)
        self.render_strings_in_dict(global_config, context, 'global')

        report_config = self.get_report_config(name)
        if report_config:
            report_variables = report_config.get('variables', {})
            self.render_strings_in_dict(report_variables, context)

        self.render_strings_in_dict(report_config, context, 'report')

        context.update({'hreport': self})

        return context

    def render(self, name):

        template_name = self.get_report_config_value(name, 'template')

        if not template_name:
            return self.run(name)

        try:
            template = self.env.get_template(template_name)
        except TemplateNotFound:
            result = 'Template %s Not Found' % template_name
            return result
        except TemplateSyntaxError:
            result = 'Template syntax error in %s' % template_name
            return result

        context = self.get_context(name)
        context['output'] = self.run(name).splitlines()

        try:
            result = template.render(context)
        except TemplateSyntaxError:
            result = 'Template raised error: "%s"' % template
        except UndefinedError as exception:
            result = 'Variable UndefinedError %s in template "%s"' % \
                (exception.message, template_name)
        return result

    def save(self, name):
        input_file = tempfile.NamedTemporaryFile(dir='.',
                                                 delete=False)
        input_file.close()

        with io.open(input_file.name, 'w',
                     encoding='utf-8') as input_file:
            input_file.write(self.render(name))
        output_file = self.get_report_config_value(name, 'filename')

        if not output_file:
            output_file = '%s.pdf' % name

        output_file = self.render_string(output_file, name)

        cmd = 'pandoc %s -t html5 -o %s' % (input_file.name,
                                            output_file)

        styling = self.get_report_config_value(name, 'styling')
        template_name = self.get_report_config_value(name,
                                                     'template')
        if styling:
            styling = os.path.join(self.cfg_templates, styling)
            styling = '--css "%s"' % styling
        elif template_name:
            styling_default = template_name.split('.')[0] + '.css'
            styling_default_path = os.path.join(self.cfg_templates,
                                                styling_default)
            if os.path.exists(styling_default_path):
                cmd = cmd + ' --css %s' % styling_default_path
        input_file.close()
        cmd_list = shlex.split(cmd)
        unicode(subprocess.check_output(cmd_list), 'utf-8')
        os.unlink(input_file.name)
        return output_file



# -*- coding: utf-8 -*-

import datetime
from jinja2.exceptions import FilterArgumentError


def datetimeformat(value, format='%H:%M / %d-%m-%Y'):
    return value.strftime(format)


def german_float(value):
    value = float(value)
    return '{:,.2f}'.format(value).replace(",", "X")\
                    .replace(".", ",").replace("X", ".")


def last_day_of_month(any_day):
    next_month = any_day.replace(day=28) + datetime.timedelta(days=4)
    return next_month - datetime.timedelta(days=next_month.day)


def substract_days(any_day,  amount):
    if amount < 0:
        new_date = any_day + datetime.timedelta(days=-1*amount)
    else:
        new_date = any_day - datetime.timedelta(days=amount)
    return new_date


def parse_multiply_last_column_input(factor, keywords):
    factor_by_keyword = {}
    if isinstance(keywords, list):
        try:
            for keyword in keywords:
                factor_by_keyword[keyword[0]] = keyword[1]
        except TypeError:
            raise FilterArgumentError
    else:
        try:
            factor = float(factor)
        except ValueError:
            raise FilterArgumentError
    return factor, factor_by_keyword


def calculate_last_column(line, value, factor, keywords):
    result = None
    if keywords:
        for keyword in keywords.keys():
            if keyword in line:
                factor = float(keywords.get(keyword))
                return '{:,.2f}'.format(value*factor)
    if not result:
        return '{:,.2f}'.format(value*factor)


def multiply_last_column(output, factor, title=None, keywords={}):
    max_length = max([len(x) for x in output])
    factor, keywords = parse_multiply_last_column_input(factor, keywords)

    extra_column = []

    # Split each line and find the last string that converts to a float
    for line in output:
        result = None
        columns = line.split(' ')
        columns.reverse()
        for cell in columns:
            try:
                cell = cell.replace(",", "")
                value = float(cell)
                result = calculate_last_column(line, value, factor, keywords)
                if result:
                    extra_column.append(result)
                    break
            except ValueError:
                value = cell

        # Append space if no float was found in line
        if not result:
            extra_column.append(" ")

    # Replace first instance with title
    if title:
        floats = [i for i, f in enumerate(extra_column) if f != " "]
        if floats:
            extra_column[floats[0]] = title

        # Special case for the percentage filter
        if title == "%":
            return extra_column

    # Append to existing output and align spacing
    max_result_value = max([len(x) for x in extra_column])
    output_with_extra_column = []

    for line, column in zip(output, extra_column):
        new_line = line + " "*(max_length-len(line)) + " " +\
                   " "*(max_result_value-len(column)) + column
        output_with_extra_column.append(new_line)

    return output_with_extra_column


def add_percentage_column(output):
    extra_column = multiply_last_column(output, 1.0, "%", {})
    floats = [i for i, f in enumerate(extra_column) if f != " "]
    for index in floats[1:]:
        extra_column[index] = float(extra_column[index].replace(',', ''))
    total = extra_column[floats[-1]]

    total_percentage = 0
    for index in floats[1:-1]:
        percentage = 100*(extra_column[index]/total)
        total_percentage += percentage
        extra_column[index] = '{:,.2f}'.format(percentage)
    extra_column[floats[-1]] = '{:,.2f}'.format(total_percentage)

    max_result_value = max([len(x) for x in extra_column])
    output_with_extra_column = []
    max_length = max([len(x) for x in output])
    for line, column in zip(output, extra_column):
        new_line = line + " "*(max_length-len(line)) + " " +\
                   " "*(max_result_value-len(column)) + column
        output_with_extra_column.append(new_line)

    return output_with_extra_column
